/*
** Apache 2 library.
** $Id: apache2_lib.c,v 1.7 2007/02/02 12:24:11 zhaozg Exp $
*/
#include <sys/types.h>

#include "httpd.h"
#include "util_filter.h"
#include "http_core.h"
#include "http_protocol.h"
#include "apr_strings.h"
#include "http_log.h"

#include "lua.h"
#include "lauxlib.h"
#include "lualib.h"

#include "apache2_lib.h"
#include "apr_network_io.h"

#include "storage.h"
#include "mod_lua.h"

/* mod_ssl.h is not safe for inclusion in 2.0, so duplicate the
* optional function declarations. */
APR_DECLARE_OPTIONAL_FN(char *, ssl_var_lookup,
						(apr_pool_t *, server_rec *,
						conn_rec *, request_rec *,
						char *));
APR_DECLARE_OPTIONAL_FN(int, ssl_is_https, (conn_rec *));


#define UP_REC(L) (*(request_rec **)lua_touserdata(L, lua_upvalueindex (1)))

typedef struct {
    const char    *t_name;
    int      t_val;
} TRANS;

static const TRANS priorities[] = {
    {"emerg",   APLOG_EMERG},
    {"alert",   APLOG_ALERT},
    {"crit",    APLOG_CRIT},
    {"error",   APLOG_ERR},
    {"warn",    APLOG_WARNING},
    {"notice",  APLOG_NOTICE},
    {"info",    APLOG_INFO},
    {"debug",   APLOG_DEBUG},
    {NULL,      -1},
};
static int default_priority = 3;

static APR_OPTIONAL_FN_TYPE(ssl_var_lookup) *lua_ssl_lookup = NULL;
static APR_OPTIONAL_FN_TYPE(ssl_is_https) *lua_is_https = NULL;


/*
** Send http header.
** Uses the request_rec defined as an upvalue.
** Receives a Lua string.
*/
static int ap2_set_content_type (lua_State *L) {
	request_rec *r = UP_REC(L);
	//ap_set_content_type (r, apr_pstrdup(r->pool,(char *)luaL_checkstring (L, 1)));
	ap_set_content_type (r, apr_pstrdup(r->pool,(char *)lua_tostring (L, 1)));
	return 0;
}

/*
** Consulting header_only field of request_rec.
** Uses the request_rec defined as an upvalue.
** Returns a Lua boolean.
*/
static int ap2_header_only (lua_State *L) {
	lua_pushboolean (L, UP_REC(L)->header_only);
	return 1;
}

static int ap2_is_https (lua_State *L) {
	lua_pushboolean (L,lua_is_https && lua_is_https(UP_REC(L)->connection));
	return 1;
}


static int ap2_ssl_lookup (lua_State *L) {
	request_rec *r = UP_REC(L);
	char* var = (char*)lua_tostring(L,1);
	char* val = NULL;

	if(lua_ssl_lookup && var && (val = lua_ssl_lookup(r->pool, r->server, r->connection, r, var)) )
		lua_pushstring(L, apr_pstrdup(r->pool,val));
	else
		lua_pushnil(L);

	return 1;
}

/*
** Consulting protocol field of request_rec.
** Uses the request_rec defined as an upvalue.
** Returns a Lua string.
*/
static int ap2_protocol (lua_State *L) {
	lua_pushstring (L, UP_REC(L)->protocol);
	return 1;
}

/*
** Consulting hostname field of request_rec.
** Uses the request_rec defined as an upvalue.
** Returns a Lua string.
*/
static int ap2_hostname (lua_State *L) {
	lua_pushstring (L, UP_REC(L)->hostname);
	return 1;
}

/*
** Consulting method field of request_rec.
** Uses the request_rec defined as an upvalue.
** Returns a Lua string.
*/
static int ap2_method (lua_State *L) {
	lua_pushstring (L, UP_REC(L)->method);
	return 1;
}

/*
** Consulting unparsed_uri field of request_rec.
** Uses the request_rec defined as an upvalue.
** Returns a Lua string.
*/
static int ap2_unparsed_uri (lua_State *L) {
	lua_pushstring (L, UP_REC(L)->unparsed_uri);
	return 1;
}

/*
** Consulting uri field of request_rec.
** Uses the request_rec defined as an upvalue.
** Returns a Lua string.
*/
static int ap2_uri (lua_State *L) {
	lua_pushstring (L, UP_REC(L)->uri);
	return 1;
}

/*
** Consulting path_info field of request_rec.
** Uses the request_rec defined as an upvalue.
** Returns a Lua string.
*/
static int ap2_path_info (lua_State *L) {
	lua_pushstring (L, UP_REC(L)->path_info);
	return 1;
}

/*
** Consulting args field of request_rec.
** Uses the request_rec defined as an upvalue.
** Returns a Lua string.
*/
static int ap2_args (lua_State *L) {
	lua_pushstring (L, UP_REC(L)->args);
	return 1;
}

/*
** Consulting handler field of request_rec.
** Uses the request_rec defined as an upvalue.
** Returns a Lua string.
*/
static int ap2_handler (lua_State *L) {
	lua_pushstring (L, UP_REC(L)->handler);
	return 1;
}

/*
 * Consulting content_type field of request_rec.
 * Uses the request_rec defined as an upvalue.
 * Returns a Lua string with the value of the Content-type header.
 */
static int ap2_get_content_type (lua_State *L) {
	lua_pushstring (L, UP_REC(L)->content_type);
	return 1;
}

/*
 * Consulting filename field of request_rec.
 * Uses the request_rec defined as an upvalue.
 * Returns a Lua string with the value of the filename field.
 */
static int ap2_filename (lua_State *L) {
	lua_pushstring (L, UP_REC(L)->filename);
	return 1;
}

static int ap2_rputs (lua_State *L) {
	int nargs = lua_gettop (L);
	int idx = 1;
	for (; nargs--; idx++) {
		size_t l;
		const char *s = luaL_checklstring(L, idx, &l);
		if (ap_rputs (s, UP_REC(L)) == -1)
			luaL_error (L, "error while writing output");
	}
	return 0;
}

/*
** Send a string to the client.
** Uses the request_rec defined as an upvalue.
** Receives a Lua string.
*/
static int ap2_rwrite (lua_State *L) {
	int nargs = lua_gettop (L);
	int idx = 1;
	for (; nargs--; idx++) {
		size_t l;
		const char *s = luaL_checklstring(L, idx, &l);
		if (ap_rwrite (s, l, UP_REC(L)) == -1)
			luaL_error (L, "error while writing output");
	}
	return 0;
}

static int pushresult (lua_State *L, int i, const char *filename) {
	if (i) {
		lua_pushboolean(L, 1);
		return 1;
	}
	else {
		char errbuf[MAX_STRING_LEN];

		lua_pushnil(L);
		if (filename)
			lua_pushfstring(L, "%s: %s", filename, 	apr_strerror(apr_get_os_error(),errbuf,MAX_STRING_LEN));
		else
			lua_pushfstring(L, "%s", apr_strerror(apr_get_os_error(),errbuf,MAX_STRING_LEN));
		lua_pushnumber(L, apr_get_os_error());
		return 3;
	}
}

/*
** Think on this, need it?
**
*/
static int ap2_write (lua_State *L) {
	int nargs = lua_gettop(L);
	int status = 1;
	int arg = 1;
	request_rec *r = UP_REC(L);

	for (; nargs--; arg++) {
		if (lua_type(L, arg) == LUA_TNUMBER) {
			/* optimization: could be done exactly as for strings */
			status = status &&
				ap_rprintf(r, LUA_NUMBER_FMT, lua_tonumber(L, arg)) > 0;
		}
		else {
			size_t l;
			const char *s = luaL_checklstring(L, arg, &l);
			status = status && (ap_rwrite(s, l, r)==l);
		}
	}
	return pushresult(L, status, r->filename);
}

/*
** Flush output buffer
**
*/
static int ap2_rflush (lua_State *L) {
	return pushresult(L, ap_rflush(UP_REC(L))>=0, NULL);
}

/*
** print
**
*/

static int ap2_print (lua_State *L) {
	int i, n;
	int status = 1;
	request_rec* r = UP_REC(L);


	n = lua_gettop(L);
	for (i=1; i<=n && status; i++) {
		size_t l;
		const char *s = luaL_checklstring(L, i, &l);
		status = status && (ap_rwrite(s, l, r)==l);
	}
	return pushresult(L, status, r->filename);
} 

/*
** Read Request Body!
**
*/
static int read_chars (lua_State *L, size_t n) {
	size_t len;
	char tmpbuf[HUGE_STRING_LEN];
	request_rec* r = UP_REC(L);

	len = HUGE_STRING_LEN;  /* try to read that much each time */

	if (len > n) 
		len = n; 

	if ( (len = ap_get_client_block(r, tmpbuf, len))> 0)
	{
		lua_pushlstring(L,tmpbuf,len);
		return 1;
	}else
	{
		ap_discard_request_body(r);
	}

	return 0;
}

static int ap2_read (lua_State *L) {
	int nargs = lua_gettop(L) ;
	int success;
	int n = 1;

	request_rec* r = UP_REC(L);

	if (r->remaining ==0 ) {
		lua_pushnil(L);  /* push nil instead */
		return 1;
	}

	if (nargs == 0) {  /* no arguments? */
		return success = read_chars(L, (size_t)~0);
	}

	luaL_checkstack(L, nargs+LUA_MINSTACK, "too many arguments");
	success = 1;

	for (n=1; n<=nargs && success; n++)
	{
		if (lua_type(L, n) == LUA_TNUMBER) {
			size_t l = (size_t)lua_tonumber(L, n);
			success = read_chars(L, l);
		}
		else 
		{
			const char *p = lua_tostring(L, n);
			luaL_argcheck(L, p && p[0] == '*' && p[1] == 'a', n, "invalid option");
			success = read_chars(L,(size_t)~0);  /* read MAX_SIZE_T chars */
		}
	}

	return n-1;
}

/*
** Binding to ap_setup_client_block.
** Uses the request_rec defined as an upvalue.
** Receives a Lua string: "REQUEST_NO_BODY", "REQUEST_CHUNKED_ERROR" or
**	"REQUEST_CHUNKED_DECHUNK".
** It returns the status code.
*/
static int ap2_setup_client_block (lua_State *L) {
	request_rec *r = UP_REC(L);
	const char *s = luaL_checklstring (L, -1, 0);
	if (strcmp (s, "REQUEST_NO_BODY") == 0)
		lua_pushnumber (L, ap_setup_client_block (r, REQUEST_NO_BODY));
	else if (strcmp (s, "REQUEST_CHUNKED_ERROR") == 0)
		lua_pushnumber (L, ap_setup_client_block (r, REQUEST_CHUNKED_ERROR));
	else if (strcmp (s, "REQUEST_CHUNKED_DECHUNK") == 0)
		lua_pushnumber (L, ap_setup_client_block (r, REQUEST_CHUNKED_DECHUNK));
	else
		lua_pushnil (L);
	return 1;
}

/*
 * Binding to ap_should_client_block.
 * Uses the request_rec defined as an upvalue.
 * Returns the status code.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
Esta funcao nao deve ser chamada mais de uma vez.
O que fazer para evitar isto?
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 */
static int ap2_should_client_block (lua_State *L) {
	lua_pushnumber (L, ap_should_client_block (UP_REC(L)));
	return 1;
}

/*
** Binding to ap_get_client_block.
** Uses the request_rec defined as an upvalue.
** Receives a number of bytes to read.
** Returns a string or nil if EOS.
*/
static int ap2_get_client_block (lua_State *L) {
	request_rec *r = UP_REC(L);
	char buffer[POST_BUFFER_SIZE+1];
	int bytesleft = luaL_optint (L, 1, POST_BUFFER_SIZE);
	int status, n;
	int count = 0;
	if (bytesleft < 0) {
		luaL_error (L, "block size must be positive");
		return 0;
	}
	while (bytesleft) {
		n = (bytesleft > POST_BUFFER_SIZE) ? POST_BUFFER_SIZE : bytesleft;
		status = ap_get_client_block (r, buffer, n);
		if (status == 0) { /* end-of-body */
			if (r->remaining > 0)
				continue; /* still has something to read */
			else
				break; /* end-of-stream */
		} else if (status == -1) { /* error or premature chunk end */
			lua_pushnil (L);
			lua_pushstring (L, "error getting client block");
			return 2;
		} else {
			bytesleft -= status;
			lua_pushlstring (L, buffer, status);
			count++;
		}
	}
	/* is this necessary? */
	if (count)
		lua_concat (L, count);
	else
		lua_pushnil (L);
	return 1;
}

/*
** Consulting remaining field of request_rec.
** Uses the request_rec defined as an upvalue.
** Return a Lua number.
*/
static int ap2_remaining (lua_State *L) {
	lua_pushnumber (L, (lua_Number)UP_REC(L)->remaining);
	return 1;
}

/*
** Binding to ap_discard_request_body.
** Uses the request_rec defined as an upvalue.
** Returns a status code.
*/
static int ap2_discard_request_body (lua_State *L) {
	lua_pushnumber (L, ap_discard_request_body (UP_REC(L)));
	return 1;
}

/*
** Consulting headers.
** Uses the request_rec defined as an upvalue.
** Receives a string with the header to search or nil.
** Returns the header value or nil if not found.
** Returns all headers on the stack.
*/
static int ap2_get_header (lua_State *L) {
	request_rec *r = UP_REC(L);
	const char *header = luaL_optstring (L, 1, NULL);
	const apr_array_header_t *hdrs_arr = apr_table_elts (r->headers_in);
	const apr_table_entry_t *hdrs = (const apr_table_entry_t *) hdrs_arr->elts;
	if (header == NULL) { /* return all headers */
		int i, count;
		luaL_checkstack (L, hdrs_arr->nelts, "too many headers");
		for (i = 0, count = 0; i < hdrs_arr->nelts; i++) {
			if (hdrs[i].key) {
				count++;
				lua_pushstring (L, hdrs[i].key);
				lua_pushliteral (L, " = ");
				lua_pushstring (L, hdrs[i].val);
				lua_concat (L, 3);
			}
		}
		return count;
	} else { /* search for a given header */
		int i;
		for (i = 0; i < hdrs_arr->nelts; i++) {
			if (!hdrs[i].key)
				continue;
			if (!strcasecmp (hdrs[i].key, header)) {
				lua_pushstring (L, hdrs[i].val);
				return 1;
			}
		}
	}
	lua_pushnil (L);
	return 1;
}

/*
** Set an output header.
** Uses the request_rec defined as an upvalue.
** Receives a string with the header's name AND
** a second string with the header's value.
** Note: apr_table_setn DOES NOT copies the key and value strings.
*/
static int ap2_set_headers_out (lua_State *L) {
	const char *header = luaL_checkstring (L, 1);
	const char *value = luaL_checkstring (L, 2);
	apr_table_set (UP_REC(L)->headers_out, header, value);
	return 0;
}


/*
** Set an output "error" header.
** Uses the request_rec defined as an upvalue.
** Receives a string with the header's name AND
** a second string with the header's value.
** Note: apr_table_setn DOES NOT copies the key and value strings.
** Note 2: this header is "printed even on errors and persist across
** internal redirects" -- from Apache's include/httpd.h
*/
static int ap2_set_err_headers_out (lua_State *L) {
	const char *header = luaL_checkstring (L, 1);
	const char *value = luaL_checkstring (L, 2);
	apr_table_set (UP_REC(L)->err_headers_out, header, value);
	return 0;
}
/*
** Add an output header.
** Uses the request_rec defined as an upvalue.
** Receives a string with the header's name AND
** a second string with the header's value.
** Note: apr_table_addn DOES NOT copies the key and value strings.
*/
static int ap2_add_headers_out (lua_State *L) {
	const char *header = luaL_checkstring (L, 1);
	const char *value = luaL_checkstring (L, 2);
	apr_table_add (UP_REC(L)->headers_out, header, value);
	return 0;
}
/*
** Add an output "error" header.
** Uses the request_rec defined as an upvalue.
** Receives a string with the header's name AND
** a second string with the header's value.
** Note: apr_table_addn DOES NOT copies the key and value strings.
** Note 2: this header is "printed even on errors and persist across
** internal redirects" -- from Apache's include/httpd.h
*/
static int ap2_add_err_headers_out (lua_State *L) {
	const char *header = luaL_checkstring (L, 1);
	const char *value = luaL_checkstring (L, 2);
	apr_table_add (UP_REC(L)->err_headers_out, header, value);
	return 0;
}
/*
                apr_table_add(rmain->err_headers_out, "Set-Cookie", cookie);
*/

/*
** Get user name.
** Uses the request_rec defined as an upvalue.
** Returns a string with the user name.
*/
static int ap2_user (lua_State *L) {
	lua_pushstring (L, UP_REC(L)->user);
	return 1;
}

/*
** Get auth type.
** Uses the request_rec defined as an upvalue.
** Returns a string with the auth type.
*/
static int ap2_auth_type (lua_State *L) {
	lua_pushstring (L, UP_REC(L)->ap_auth_type);
	return 1;
}

/*
** Bind to get_remote_host.
** Uses the request_rec defined as an upvalue.
** Returns a string with the remote host name.
*/
static int ap2_get_remote_host (lua_State *L) {
	request_rec *r = UP_REC(L);
	lua_pushstring (L, ap_get_remote_host (r->connection, r->per_dir_config, REMOTE_HOST, NULL));
	return 1;
}

/*
** Get remote logname.
** Uses the request_rec defined as an upvalue.
** Returns a string with the remote logname.
*/
static int ap2_get_remote_logname (lua_State *L) {
	lua_pushstring (L, ap_get_remote_logname(UP_REC(L)));
	return 1;
}

/*
** Bind to get_server_version.
** Uses the request_rec defined as an upvalue.
** Returns a string with the server software version.
*/
static int ap2_get_server_version (lua_State *L) {
	lua_pushstring (L, ap_get_server_version());
	return 1;
}

/*
** Bind to get_server_name.
** Uses the request_rec defined as an upvalue.
** Returns a string with the server name.
*/
static int ap2_get_server_name (lua_State *L) {
	request_rec *r = UP_REC(L);
	lua_pushstring (L, ap_escape_html (r->pool, ap_get_server_name(r)));
	return 1;
}

/*
** Bind to get_server_port.
** Uses the request_rec defined as an upvalue.
** Returns a string with the server name.
*/
static int ap2_get_server_port (lua_State *L) {
	request_rec *r = UP_REC(L);
	lua_pushnumber (L, ap_get_server_port(r));
	return 1;
}

/*
** Obtain the remote IP address.
** Uses the request_rec defined as an upvalue.
** Returns a string with the remote IP address.
*/
static int ap2_conn_remote_addr (lua_State *L) {
	request_rec *r = UP_REC(L);
	lua_pushstring (L, r->connection->remote_ip);
	return 1;
}

/*
** Obtain the remote port.
** Uses the request_rec defined as an upvalue.
** Returns a string with the remote port.
*/
static int ap2_conn_remote_port (lua_State *L) {
	apr_port_t rport;
	request_rec *r = UP_REC(L);
	//apr_sockaddr_port_get (&rport, r->connection->remote_addr);
	rport = r->connection->remote_addr->port;
	lua_pushstring (L, apr_itoa (r->pool, rport));
	return 1;
}


/*
** Convert a string to a level code.
*/
static int str2level (const char *s) {
	int i;
	for (i = 0; priorities[i].t_name; i++)
		if (strcmp (s, priorities[i].t_name) == 0)
			return priorities[i].t_val;
	return priorities[default_priority].t_val;
}


/*
** Send a string to the error log.
** Receives a string with a message, an optional level string (valid strings
** are "emerg", "alert", "crit", "error", "warn", "notice", "info", "debug"),
** and an optional status number.
*/
static int ap2_log_error (lua_State *L) {
	request_rec *r = UP_REC(L);
	const char *message = luaL_checkstring (L, 1);
	int level = str2level (luaL_optstring (L, 2, ""));
	apr_status_t status = (apr_status_t) luaL_optnumber (L, 3, APR_SUCCESS);
	ap_log_error (APLOG_MARK, level, status, r->server, message);
	return 0;
}


/*
** Retrieves the process identification (PID) and its parent's.
*/
#if !defined(getpid)
#if defined(_getpid)
#define getpid _getpid
#endif
#endif
static int ap2_pid (lua_State *L) {
	lua_pushnumber (L, (lua_Number)getpid ());
#if defined(getppid)
	lua_pushnumber (L, (lua_Number)getppid ());
#else
	lua_pushnil (L);
#endif
	return 2;
}

/**************************************************/
#define setstr(key,value)\
	lua_pushstring(L,key);\
	if (value) lua_pushstring(L,value); else lua_pushnil(L);\
	lua_settable(L,-3)

#define setnum(key,value)\
	lua_pushstring(L,key);\
	lua_pushnumber(L,(double)value);\
	lua_settable(L,-3)

#define setnum2(key,value)\
	lua_pushnumber(L,key);\
	lua_pushnumber(L,(double)value);\
	lua_settable(L,-3)

static int ap2_env(lua_State *L)
{
	int i;
	int nargs = lua_gettop(L) ;

	apr_table_t *e = NULL;

	const apr_array_header_t *env_arr ;
	const apr_table_entry_t *elts ;

	request_rec* r = UP_REC(L);

	e = r->subprocess_env;
	env_arr = apr_table_elts(e);
	elts = (const apr_table_entry_t *) env_arr->elts;

	if(nargs==0)
	{
		lua_newtable(L);  /* create it */
		for (i = 0; i < env_arr->nelts; ++i) {
			if (!elts[i].key)
				continue;

			setstr(elts[i].key,elts[i].val);
		}
		return 1;
	}
	for(i=1; i<=nargs; i++)
	{
		const char *key = lua_tostring(L,i);
		const char *val = apr_table_get(e,key);
		if(val)
			lua_pushstring(L,val);
		else
			lua_pushnil(L);
	}
	return nargs;
}

static int ap2_header (lua_State *L) {
	int nargs = lua_gettop(L);
	const char* key = NULL;
	const char* val = NULL;

	request_rec* r = UP_REC(L);

	if(nargs==0)
	{
		int i;
		const apr_array_header_t *header_arr = apr_table_elts(r->headers_in);
		const apr_table_entry_t *elts = (const apr_table_entry_t *) header_arr->elts;

		lua_newtable(L);  /* create it */
		for (i = 0; i < header_arr->nelts; ++i) {
			if (!elts[i].key)
				continue;

			setstr(elts[i].key,elts[i].val);
		}
		return 1;
	}else if(nargs==1)    /* get header value */
	{
		key = lua_tostring(L, 1);
		val = apr_table_get(r->headers_in,key);
		if(val)
			lua_pushstring(L,val);
		else
			lua_pushnil(L);
		return 1;
	}else if(nargs==2)    /* set header value */
	{
		key = lua_tostring(L,1);
		val = lua_tostring(L,2);
		if(apr_strnatcasecmp(key, "Content-Type")==0)
			ap_set_content_type(r,apr_pstrdup(r->pool, val));

		apr_table_set(r->headers_out,key,val);
		return 0;
	}
	return 0;
}



static int ap2_status(lua_State *L)
{
	int nargs = lua_gettop(L);

	request_rec *r = UP_REC(L);
	if(nargs==0)
	{
		lua_pushnumber(L,r->status);
		return 1;
	}else if(nargs==1)
	{
		r->status = (int)lua_tointeger(L,1);
		
	}
	return 0;
}



static int ap2_storage(lua_State *L)
{
	server_rec *s = NULL;
	BOOL rc = FALSE;
	int nargs = lua_gettop(L);
	request_rec *r = UP_REC(L);

	s = r->server;
	s = r->connection->base_server;

	if(nargs==0)
	{
		//tableʽе
		LUAConfigRecord *mc = ThisConfigRecord(s);
		//LUASrvConfigRec *sc = ThisSrvConfig(s);

		lua_newtable(L);  /* create it */

		if (mc->nStorageMode == STORAGE_SCMODE_DBM)
		{
			apr_dbm_t *dbm;
			apr_datum_t dbmkey;
			apr_datum_t dbmval;
			apr_pool_t *p;
			time_t tExpiresAt;
			apr_datum_t *keylist;
			int keyidx;
			apr_status_t rv;

			/*
			* Here we have to be very carefully: Not all DBM libraries are
			* smart enough to allow one to iterate over the elements and at the
			* same time delete expired ones. Some of them get totally crazy
			* while others have no problems. So we have to do it the slower but
			* more safe way: we first iterate over all elements and remember
			* those which have to be expired. Then in a second pass we delete
			* all those expired elements. Additionally we reopen the DBM file
			* to be really safe in state.
			*/

#define KEYMAX 1024

			storage_mutex_on(s);
			for (;;) {
				/* allocate the key array in a memory sub pool */
#if AP_SERVER_MINORVERSION_NUMBER>=2
				apr_pool_create_ex(&p, mc->pPool, NULL,NULL);
#elif AP_SERVER_MINORVERSION_NUMBER==0
				apr_pool_sub_make(&p, mc->pPool, NULL);
#endif

				if (p == NULL)
					break;
				if ((keylist = apr_palloc(p, sizeof(dbmkey)*KEYMAX)) == NULL) {
					apr_pool_destroy(p);
					break;
				}

				/* pass 1: scan DBM database */
				keyidx = 0;
				if ((rv = apr_dbm_open(&dbm, mc->szStorageDataFile, 
					APR_DBM_RWCREATE,STORAGE_DBM_FILE_MODE,
					p)) != APR_SUCCESS) {
						ap_log_error(APLOG_MARK, APLOG_ERR, rv, s,
							"Cannot open LUASessionCache DBM file `%s' for "
							"scanning",
							mc->szStorageDataFile);
						apr_pool_destroy(p);
						break;
					}
					apr_dbm_firstkey(dbm, &dbmkey);
					while (dbmkey.dptr != NULL) {
						apr_dbm_fetch(dbm, dbmkey, &dbmval);
						memcpy(&tExpiresAt, dbmval.dptr, sizeof(time_t));
						setnum(dbmkey.dptr,tExpiresAt);
 						apr_dbm_nextkey(dbm, &dbmkey);
					}
					apr_dbm_close(dbm);
			}
			storage_mutex_off(s);
		}
		else if (mc->nStorageMode == STORAGE_SCMODE_SHMHT)
		{
			//reference expire function

			table_linear_t iterator;
			time_t tExpiresAt;
			void *vpKey;
			void *vpKeyThis;
			void *vpData;
			int nKey;
			int nKeyThis;
			int nData;
			int rc;
			char *p;

			storage_mutex_on(s);
			if (table_first_r(mc->tStorageDataTable, &iterator,
				&vpKey, &nKey, &vpData, &nData) == TABLE_ERROR_NONE) {
					do {
						memcpy(&tExpiresAt, vpData, sizeof(time_t));

						vpKeyThis = vpKey;
						nKeyThis  = nKey;
						p = (char*)vpKeyThis;
						p[nKey]='\0';
						setnum(vpKeyThis,tExpiresAt);
						rc = table_next_r(mc->tStorageDataTable, &iterator,
							&vpKey, &nKey, &vpData, &nData);
					} while (rc == TABLE_ERROR_NONE); 
					/* (vpKeyThis != vpKey) && (nKeyThis != nKey) */
				}
				storage_mutex_off(s);
		} if (mc->nStorageMode == STORAGE_SCMODE_SHMCB)
		{
			//reference status function
			SHMCBHeader *header;
			SHMCBQueue queue;
			SHMCBCache cache;
			SHMCBIndex *idx;
			unsigned int loop;
			unsigned int i;

			storage_mutex_on(s);
			ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, "inside storage_shmcb_iterator");

			/* Get the header structure. */
			storage_shmcb_get_header((void *) mc->tStorageDataTable, &header);

			/* It may seem strange to grab "now" at this point, but in theory
			 * we should never have a negative threshold but grabbing "now" after
			 * the loop (which performs expiries) could allow that chance. */
			for (loop = 0; loop <= header->division_mask; loop++) {
				if (storage_shmcb_get_division(header, &queue, &cache, loop)) {
					storage_shmcb_expire_division(s, &queue, &cache);
					for(i = storage_shmcb_get_safe_uint(queue.first_pos); 
						i < storage_shmcb_get_safe_uint(queue.pos_count);
						i++
						)
					{
						idx = storage_shmcb_get_index(&queue,storage_shmcb_get_safe_uint(&i));
						if(idx!=NULL && !idx->removed)
						{
							setnum2(idx->key,storage_shmcb_get_safe_time(&(idx->expires)));
						}
					}
				}
			}
			storage_mutex_off(s);
		}

		return 1;
	}else if(nargs==1)
	{
		//ز
		int l;
		const char *key = lua_tostring(L,1);
		char *val = (char*)storage_retrieve(s,(UCHAR*)key,lua_strlen(L,1),&l);
		if(val==NULL)
			lua_pushnil(L);
		else
		{
			lua_pushlstring(L,val,l);
			free(val);
		}
		return 1;

	}else if(nargs==2)
	{
		//趨key=valueʽ
		LUASrvConfigRec *sc = ThisSrvConfig(s);

		const char* key = lua_tostring(L,1);
		const char* val = lua_tostring(L,2);
		time_t expiry = time(NULL);

		if (lua_type(L,2)==LUA_TNIL)
		{
			storage_remove(s,(UCHAR*)key,lua_strlen(L,1));
			return 0;
		}

		expiry += sc->storage_timeout;
		rc = storage_store(s,(UCHAR*)key,lua_strlen(L,1),expiry,(void*)val,lua_strlen(L,2));
		return 0;
	}else if(nargs==3)
	{
		//趨key=valueʽ,ֵΪʱʱ
		//LUASrvConfigRec *sc = ThisSrvConfig(s);

		const char* key = lua_tostring(L,1);
		const char* val = lua_tostring(L,2);
		time_t expiry = (time_t)lua_tonumber(L,3);
		rc = storage_store(s,(UCHAR*)key,lua_strlen(L,1),expiry,(void*)val,lua_strlen(L,2));
		return 0;
	}
	return 0;
}

#undef setstr
#undef setnum

/*
** Binding functions
*/
static const luaL_reg ap2_lib[] = {
	{ "is_https",             ap2_is_https },
	{ "ssl_lookup",           ap2_ssl_lookup },
	{ "set_content_type",     ap2_set_content_type },
	{ "header_only",          ap2_header_only },
	{ "protocol",             ap2_protocol },
	{ "hostname",             ap2_hostname },
	{ "method",               ap2_method },
	{ "unparsed_uri",         ap2_unparsed_uri },
	{ "uri",                  ap2_uri },
	{ "path_info",            ap2_path_info },
	{ "args",                 ap2_args },
	{ "handler",              ap2_handler },
	{ "get_content_type",     ap2_get_content_type },
	{ "filename",             ap2_filename },
	{ "print",                ap2_print},
	{ "rputs",                ap2_rputs },
	{ "rwrite",               ap2_rwrite },
	{ "write",                ap2_write},
	{ "read",                 ap2_read},
	{ "rflush",				  ap2_rflush},
	{ "setup_client_block",   ap2_setup_client_block },
	{ "should_client_block",  ap2_should_client_block },
	{ "get_client_block",     ap2_get_client_block },
	{ "remaining",            ap2_remaining },
	{ "discard_request_body", ap2_discard_request_body },
	{ "get_header",           ap2_get_header },
	{ "set_headers_out",      ap2_set_headers_out },
	{ "set_err_headers_out",  ap2_set_err_headers_out },
	{ "add_headers_out",      ap2_add_headers_out },
	{ "add_err_headers_out",  ap2_add_err_headers_out },
	{ "user",                 ap2_user },
	{ "auth_type",            ap2_auth_type },
	{ "get_remote_host",      ap2_get_remote_host },
	{ "get_remote_logname",   ap2_get_remote_logname },
	{ "get_server_version",   ap2_get_server_version },
	{ "get_server_name",      ap2_get_server_name },
	{ "get_server_port",      ap2_get_server_port },
	{ "conn_remote_addr",     ap2_conn_remote_addr },
	{ "conn_remote_port",     ap2_conn_remote_port },
	{ "log_error",            ap2_log_error },
	{ "pid",                  ap2_pid },

	{ "header", ap2_header},
	{ "status", ap2_status},
	{ "env",    ap2_env},
	{ "storage",ap2_storage},

	{ NULL, NULL }
};

typedef struct {
	const char *name;
	int val;
} ap2_constants;

static ap2_constants tab_consts[] = {
	{ "OK",                         OK },
	{ "DECLINED",                   DECLINED },
	{ "HTTP_BAD_REQUEST",           HTTP_BAD_REQUEST },
	{ "HTTP_UNAUTHORIZED",          HTTP_UNAUTHORIZED },
	{ "HTTP_FORBIDDEN",             HTTP_FORBIDDEN },
	{ "HTTP_NOT_FOUND",             HTTP_NOT_FOUND },
	{ "HTTP_METHOD_NOT_ALLOWED",    HTTP_METHOD_NOT_ALLOWED },
	{ "HTTP_INTERNAL_SERVER_ERROR", HTTP_INTERNAL_SERVER_ERROR },
	{ "HTTP_MOVED_TEMPORARILY",     HTTP_MOVED_TEMPORARILY },
	{ "AP_FILTER_ERROR",            AP_FILTER_ERROR },
	{ NULL, 0 }
};

/*
** Set apachelib constants at the table on top of the stack.
*/
static void ap2_define_constants (lua_State *L, ap2_constants tab[]) {
	int i;
	for (i = 0; tab[i].name != NULL; i++) {
		lua_pushstring (L, tab[i].name);
		lua_pushnumber (L, tab[i].val);
		lua_settable (L, -3);
	}
}


void stack_dump (lua_State *L);
/*
** Create a full userdata to store the request record.
** Return a pointer to the new object AND pushes the userdata on top
** of the stack.
 */
 static request_rec **ap2_new_request (lua_State *L) {
 	request_rec **r = (request_rec **)lua_newuserdata(L,sizeof(request_rec *));
 	*r = NULL;
 	return r;
 }

/*
** Register binding functions and the constants.
** All of them use the request_rec as an upvalue.
** Leaves the table on top of the stack.
*/
int luaopen_ap2 (lua_State *L) {
	request_rec **r = ap2_new_request(L);
	if(r==NULL)
		return 0;

	luaL_openlib (L, APACHE_LIBNAME, ap2_lib, 1); /* leaves the table */
	ap2_define_constants (L, tab_consts);
	lua_pushliteral (L, APACHE_LIBNAME);
	lua_gettable (L, LUA_GLOBALSINDEX); /* stack = _G.ap */
	lua_pushliteral (L, "set_content_type");
	lua_gettable (L, -2); /* stack = ap.set_content_type */

	lua_ssl_lookup = APR_RETRIEVE_OPTIONAL_FN(ssl_var_lookup);
	lua_is_https = APR_RETRIEVE_OPTIONAL_FN(ssl_is_https);

	return 1;
}


/*
** Modify the pointer in the library's upvalue.
*/
void ap2_set_request_rec (lua_State *L, request_rec *r) {
	lua_settop(L,3); //Reset Stack

	if (lua_getupvalue (L, -1, 1) == NULL)
 			luaL_error (L, "Error while changing upvalue of Apache Library");
	*(request_rec **)lua_touserdata (L, -1) = r;
}

