/* Copyright 2001-2005 The Apache Software Foundation or its licensors, as
 * applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*                      _             _
 *  storage_dbm.c
 *  Session Cache via DBM
 */

#include "storage.h"

void storage_dbm_init(server_rec *s, apr_pool_t *p)
{
	LUAConfigRecord *mc = ThisConfigRecord(s);
    apr_dbm_t *dbm;
    apr_status_t rv;

    /* for the DBM we need the data file */
    if (mc->szStorageDataFile == NULL) {
        ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
                     "LUASessionCache required");
        storage_die();
    }

    /* open it once to create it and to make sure it _can_ be created */
    storage_mutex_on(s);
    if ((rv = apr_dbm_open(&dbm, mc->szStorageDataFile,
	    APR_DBM_RWCREATE, STORAGE_DBM_FILE_MODE, mc->pPool)) != APR_SUCCESS) {
        ap_log_error(APLOG_MARK, APLOG_ERR, rv, s,
                     "Cannot create LUASessionCache DBM file `%s'",
                     mc->szStorageDataFile);
        storage_mutex_off(s);
        return;
    }
    apr_dbm_close(dbm);

#if !defined(OS2) && !defined(WIN32) && !defined(BEOS) && !defined(NETWARE)
    /*
     * We have to make sure the Apache child processes have access to
     * the DBM file. But because there are brain-dead platforms where we
     * cannot exactly determine the suffixes we try all possibilities.
     */
    if (geteuid() == 0 /* is superuser */) {
        chown(mc->szStorageDataFile, unixd_config.user_id, -1 /* no gid change */);
        if (chown(apr_pstrcat(p, mc->szStorageDataFile, STORAGE_DBM_FILE_SUFFIX_DIR, NULL),
                  unixd_config.user_id, -1) == -1) {
            if (chown(apr_pstrcat(p, mc->szStorageDataFile, ".db", NULL),
                      unixd_config.user_id, -1) == -1)
                chown(apr_pstrcat(p, mc->szStorageDataFile, ".dir", NULL),
                      unixd_config.user_id, -1);
        }
        if (chown(apr_pstrcat(p, mc->szStorageDataFile, STORAGE_DBM_FILE_SUFFIX_PAG, NULL),
                  unixd_config.user_id, -1) == -1) {
            if (chown(apr_pstrcat(p, mc->szStorageDataFile, ".db", NULL),
                      unixd_config.user_id, -1) == -1)
                chown(apr_pstrcat(p, mc->szStorageDataFile, ".pag", NULL),
                      unixd_config.user_id, -1);
        }
    }
#endif
    storage_mutex_off(s);
    storage_dbm_expire(s);
    return;
}

void storage_dbm_kill(server_rec *s)
{
    LUAConfigRecord *mc = ThisConfigRecord(s);
    apr_pool_t *p;

#if AP_SERVER_MINORVERSION_NUMBER>=2
	apr_pool_create_ex(&p, mc->pPool, NULL,NULL);
#elif AP_SERVER_MINORVERSION_NUMBER==0
	apr_pool_sub_make(&p, mc->pPool, NULL);
#endif

    if (p != NULL) {
        /* the correct way */
        unlink(apr_pstrcat(p, mc->szStorageDataFile, STORAGE_DBM_FILE_SUFFIX_DIR, NULL));
        unlink(apr_pstrcat(p, mc->szStorageDataFile, STORAGE_DBM_FILE_SUFFIX_PAG, NULL));
        /* the additional ways to be sure */
        unlink(apr_pstrcat(p, mc->szStorageDataFile, ".dir", NULL));
        unlink(apr_pstrcat(p, mc->szStorageDataFile, ".pag", NULL));
        unlink(apr_pstrcat(p, mc->szStorageDataFile, ".db", NULL));
        unlink(mc->szStorageDataFile);
        apr_pool_destroy(p);
    }
    return;
}

BOOL storage_dbm_store(server_rec *s, UCHAR *id, int idlen, time_t expiry, void *pdata, int ndata)
{
    LUAConfigRecord *mc = ThisConfigRecord(s);
    apr_dbm_t *dbm;
    apr_datum_t dbmkey;
    apr_datum_t dbmval;
    apr_status_t rv;


    /* be careful: do not try to store too much bytes in a DBM file! */
#ifdef PAIRMAX
    if ((idlen + ndata) >= PAIRMAX) {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
                 "data size too large for DBM storage cache: %d >= %d",
                 (idlen + ndata), PAIRMAX);
        return FALSE;
    }
#else
    if ((idlen + ndata) >= 950 /* at least less than approx. 1KB */) {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
                 "data size too large for DBM storage cache: %d >= %d",
                 (idlen + ndata), 950);
        return FALSE;
    }
#endif

    /* create DBM key */
    dbmkey.dptr  = (char *)id;
    dbmkey.dsize = idlen;

    /* create DBM value */
    dbmval.dsize = sizeof(time_t) + ndata;
    dbmval.dptr  = (char *)malloc(dbmval.dsize);
    if (dbmval.dptr == NULL) {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
                 "malloc error creating DBM value");
        return FALSE;
    }
    memcpy((char *)dbmval.dptr, &expiry, sizeof(time_t));
    memcpy((char *)dbmval.dptr+sizeof(time_t), pdata, ndata);

    /* and store it to the DBM file */
    storage_mutex_on(s);
    if ((rv = apr_dbm_open(&dbm, mc->szStorageDataFile,
	    APR_DBM_RWCREATE, STORAGE_DBM_FILE_MODE, mc->pPool)) != APR_SUCCESS) {
        ap_log_error(APLOG_MARK, APLOG_ERR, rv, s,
                     "Cannot open LUASessionCache DBM file `%s' for writing "
                     "(store)",
                     mc->szStorageDataFile);
        storage_mutex_off(s);
        free(dbmval.dptr);
        return FALSE;
    }
    if ((rv = apr_dbm_store(dbm, dbmkey, dbmval)) != APR_SUCCESS) {
        ap_log_error(APLOG_MARK, APLOG_ERR, rv, s,
                     "Cannot store storage to DBM file `%s'",
                     mc->szStorageDataFile);
        apr_dbm_close(dbm);
        storage_mutex_off(s);
        free(dbmval.dptr);
        return FALSE;
    }
    apr_dbm_close(dbm);
    storage_mutex_off(s);

    /* free temporary buffers */
    free(dbmval.dptr);

    /* allow the regular expiring to occur */
    storage_dbm_expire(s);

    return TRUE;
}

void *storage_dbm_retrieve(server_rec *s, UCHAR *id, int idlen, int* ndata)
{
    LUAConfigRecord *mc = ThisConfigRecord(s);
    apr_dbm_t *dbm;
    apr_datum_t dbmkey;
    apr_datum_t dbmval;
	void *pdata;
    time_t expiry;
    time_t now;
    apr_status_t rc;

    /* allow the regular expiring to occur */
    storage_dbm_expire(s);

    /* create DBM key and values */
    dbmkey.dptr  = (char *)id;
    dbmkey.dsize = idlen;

    /* and fetch it from the DBM file 
     * XXX: Should we open the dbm against r->pool so the cleanup will
     * do the apr_dbm_close? This would make the code a bit cleaner.
     */
    storage_mutex_on(s);
    if ((rc = apr_dbm_open(&dbm, mc->szStorageDataFile,
	    APR_DBM_RWCREATE, STORAGE_DBM_FILE_MODE, mc->pPool)) != APR_SUCCESS) {
        ap_log_error(APLOG_MARK, APLOG_ERR, rc, s,
                     "Cannot open LUASessionCache DBM file `%s' for reading "
                     "(fetch)",
                     mc->szStorageDataFile);
        storage_mutex_off(s);
        return NULL;
    }
    rc = apr_dbm_fetch(dbm, dbmkey, &dbmval);
    if (rc != APR_SUCCESS) {
        apr_dbm_close(dbm);
        storage_mutex_off(s);
        return NULL;
    }
    if (dbmval.dptr == NULL || dbmval.dsize <= sizeof(time_t)) {
        apr_dbm_close(dbm);
        storage_mutex_off(s);
        return NULL;
    }

    /* parse resulting data */
    *ndata = dbmval.dsize-sizeof(time_t);
    pdata = malloc(*ndata);
    if (pdata == NULL) {
        apr_dbm_close(dbm);
        storage_mutex_off(s);
        return NULL;
    }
    /* Cast needed, ucpData may be const */
    memcpy((unsigned char *)pdata, 
           (char *)dbmval.dptr + sizeof(time_t), *ndata);
    memcpy(&expiry, dbmval.dptr, sizeof(time_t));

    apr_dbm_close(dbm);
    storage_mutex_off(s);

    /* make sure the stuff is still not expired */
    now = time(NULL);
    if (expiry <= now) {
        storage_dbm_remove(s, id, idlen);
        return NULL;
    }

    return pdata;
}

void storage_dbm_remove(server_rec *s, UCHAR *id, int idlen)
{
    LUAConfigRecord *mc = ThisConfigRecord(s);
    apr_dbm_t *dbm;
    apr_datum_t dbmkey;
    apr_status_t rv;

    /* create DBM key and values */
    dbmkey.dptr  = (char *)id;
    dbmkey.dsize = idlen;

    /* and delete it from the DBM file */
    storage_mutex_on(s);
    if ((rv = apr_dbm_open(&dbm, mc->szStorageDataFile,
	    APR_DBM_RWCREATE, STORAGE_DBM_FILE_MODE, mc->pPool)) != APR_SUCCESS) {
        ap_log_error(APLOG_MARK, APLOG_ERR, rv, s,
                     "Cannot open LUASessionCache DBM file `%s' for writing "
                     "(delete)",
                     mc->szStorageDataFile);
        storage_mutex_off(s);
        return;
    }
    apr_dbm_delete(dbm, dbmkey);
    apr_dbm_close(dbm);
    storage_mutex_off(s);

    return;
}

void storage_dbm_expire(server_rec *s)
{
    LUAConfigRecord *mc = ThisConfigRecord(s);
	LUASrvConfigRec *sc = ThisSrvConfig(s);

    static time_t tLast = 0;
    apr_dbm_t *dbm;
    apr_datum_t dbmkey;
    apr_datum_t dbmval;
    apr_pool_t *p;
    time_t tExpiresAt;
    int nElements = 0;
    int nDeleted = 0;
    int bDelete;
    apr_datum_t *keylist;
    int keyidx;
    int i;
    time_t tNow;
    apr_status_t rv;

    /*
     * make sure the expiration for still not-accessed storage
     * cache entries is done only from time to time
     */
    tNow = time(NULL);
    if (tNow < tLast+sc->storage_timeout)
        return;
    tLast = tNow;

    /*
     * Here we have to be very carefully: Not all DBM libraries are
     * smart enough to allow one to iterate over the elements and at the
     * same time delete expired ones. Some of them get totally crazy
     * while others have no problems. So we have to do it the slower but
     * more safe way: we first iterate over all elements and remember
     * those which have to be expired. Then in a second pass we delete
     * all those expired elements. Additionally we reopen the DBM file
     * to be really safe in state.
     */

#define KEYMAX 1024

    storage_mutex_on(s);
    for (;;) {
        /* allocate the key array in a memory sub pool */
#if AP_SERVER_MINORVERSION_NUMBER>=2
		apr_pool_create_ex(&p, mc->pPool, NULL,NULL);
#elif AP_SERVER_MINORVERSION_NUMBER==0
		apr_pool_sub_make(&p, mc->pPool, NULL);
#endif

        if (p == NULL)
            break;
        if ((keylist = apr_palloc(p, sizeof(dbmkey)*KEYMAX)) == NULL) {
            apr_pool_destroy(p);
            break;
        }

        /* pass 1: scan DBM database */
        keyidx = 0;
        if ((rv = apr_dbm_open(&dbm, mc->szStorageDataFile, 
                               APR_DBM_RWCREATE,STORAGE_DBM_FILE_MODE,
                               p)) != APR_SUCCESS) {
            ap_log_error(APLOG_MARK, APLOG_ERR, rv, s,
                         "Cannot open LUASessionCache DBM file `%s' for "
                         "scanning",
                         mc->szStorageDataFile);
            apr_pool_destroy(p);
            break;
        }
        apr_dbm_firstkey(dbm, &dbmkey);
        while (dbmkey.dptr != NULL) {
            nElements++;
            bDelete = FALSE;
            apr_dbm_fetch(dbm, dbmkey, &dbmval);
            if (dbmval.dsize <= sizeof(time_t) || dbmval.dptr == NULL)
                bDelete = TRUE;
            else {
                memcpy(&tExpiresAt, dbmval.dptr, sizeof(time_t));
                if (tExpiresAt <= tNow)
                    bDelete = TRUE;
            }
            if (bDelete) {
                if ((keylist[keyidx].dptr = apr_palloc(p, dbmkey.dsize)) != NULL) {
                    memcpy(keylist[keyidx].dptr, dbmkey.dptr, dbmkey.dsize);
                    keylist[keyidx].dsize = dbmkey.dsize;
                    keyidx++;
                    if (keyidx == KEYMAX)
                        break;
                }
            }
            apr_dbm_nextkey(dbm, &dbmkey);
        }
        apr_dbm_close(dbm);

        /* pass 2: delete expired elements */
        if (apr_dbm_open(&dbm, mc->szStorageDataFile,
		APR_DBM_RWCREATE,STORAGE_DBM_FILE_MODE, p) != APR_SUCCESS) {
            ap_log_error(APLOG_MARK, APLOG_ERR, rv, s,
                         "Cannot re-open LUASessionCache DBM file `%s' for "
                         "expiring",
                         mc->szStorageDataFile);
            apr_pool_destroy(p);
            break;
        }
        for (i = 0; i < keyidx; i++) {
            apr_dbm_delete(dbm, keylist[i]);
            nDeleted++;
        }
        apr_dbm_close(dbm);

        /* destroy temporary pool */
        apr_pool_destroy(p);

        if (keyidx < KEYMAX)
            break;
    }
    storage_mutex_off(s);

    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
                 "Inter-Process Session Cache (DBM) Expiry: "
                 "old: %d, new: %d, removed: %d",
                 nElements, nElements-nDeleted, nDeleted);
    return;
}

void storage_dbm_status(server_rec *s, apr_pool_t *p, void (*func)(char *, void *), void *arg)
{
    LUAConfigRecord *mc = ThisConfigRecord(s);
    apr_dbm_t *dbm;
    apr_datum_t dbmkey;
    apr_datum_t dbmval;
    int nElem;
    int nSize;
    int nAverage;
    apr_status_t rv;

    nElem = 0;
    nSize = 0;
    storage_mutex_on(s);
    /*
     * XXX - Check what pool is to be used - TBD
     */
    if ((rv = apr_dbm_open(&dbm, mc->szStorageDataFile,
	                       APR_DBM_RWCREATE, STORAGE_DBM_FILE_MODE,
                           mc->pPool)) != APR_SUCCESS) {
        ap_log_error(APLOG_MARK, APLOG_ERR, rv, s,
                     "Cannot open LUASessionCache DBM file `%s' for status "
                     "retrival",
                     mc->szStorageDataFile);
        storage_mutex_off(s);
        return;
    }
    /*
     * XXX - Check the return value of apr_dbm_firstkey, apr_dbm_fetch - TBD
     */
    apr_dbm_firstkey(dbm, &dbmkey);
    for ( ; dbmkey.dptr != NULL; apr_dbm_nextkey(dbm, &dbmkey)) {
        apr_dbm_fetch(dbm, dbmkey, &dbmval);
        if (dbmval.dptr == NULL)
            continue;
        nElem += 1;
        nSize += dbmval.dsize;
    }
    apr_dbm_close(dbm);
    storage_mutex_off(s);
    if (nSize > 0 && nElem > 0)
        nAverage = nSize / nElem;
    else
        nAverage = 0;
    func(apr_psprintf(p, "cache type: <b>DBM</b>, maximum size: <b>unlimited</b><br>"), arg);
    func(apr_psprintf(p, "current storages: <b>%d</b>, current size: <b>%d</b> bytes<br>", nElem, nSize), arg);
    func(apr_psprintf(p, "average storage size: <b>%d</b> bytes<br>", nAverage), arg);
    return;
}

