#include "storage.h"
#include <assert.h>
/*
* static unsigned int hash
*
* DESCRIPTION:
*
* Hash a variable-length key into a 32-bit value.  Every bit of the
* key affects every bit of the return value.  Every 1-bit and 2-bit
* delta achieves avalanche.  About (6 * len + 35) instructions.  The
* best hash table sizes are powers of 2.  There is no need to use mod
* (sooo slow!).  If you need less than 32 bits, use a bitmask.  For
* example, if you need only 10 bits, do h = (h & hashmask(10)); In
* which case, the hash table should have hashsize(10) elements.
*
* By Bob Jenkins, 1996.  bob_jenkins@compuserve.com.  You may use
* this code any way you wish, private, educational, or commercial.
* It's free.  See
* http://ourworld.compuserve.com/homepages/bob_jenkins/evahash.htm
* Use for hash table lookup, or anything where one collision in 2^^32
* is acceptable.  Do NOT use for cryptographic purposes.
*
* RETURNS:
*
* Returns a 32-bit hash value.
*
* ARGUMENTS:
*
* key - Key (the unaligned variable-length array of bytes) that we
* are hashing.
*
* length - Length of the key in bytes.
*
* init_val - Initialization value of the hash if you need to hash a
* number of strings together.  For instance, if you are hashing N
* strings (unsigned char **)keys, do it like this:
*
* for (i=0, h=0; i<N; ++i) h = hash( keys[i], len[i], h);
*/
unsigned int hash(const unsigned char *key,
						 const unsigned int length,
						 const unsigned int init_val)
{
	const unsigned char *key_p = key;
	unsigned int a, b, c, len;

	/* set up the internal state */
	a = 0x9e3779b9;             /* the golden ratio; an arbitrary value */
	b = 0x9e3779b9;
	c = init_val;               /* the previous hash value */

	/* handle most of the key */
	for (len = length; len >= 12; len -= 12) {
		a += (key_p[0]
		+ ((unsigned long) key_p[1] << 8)
			+ ((unsigned long) key_p[2] << 16)
			+ ((unsigned long) key_p[3] << 24));
		b += (key_p[4]
		+ ((unsigned long) key_p[5] << 8)
			+ ((unsigned long) key_p[6] << 16)
			+ ((unsigned long) key_p[7] << 24));
		c += (key_p[8]
		+ ((unsigned long) key_p[9] << 8)
			+ ((unsigned long) key_p[10] << 16)
			+ ((unsigned long) key_p[11] << 24));
		HASH_MIX(a, b, c);
		key_p += 12;
	}

	c += length;

	/* all the case statements fall through to the next */
	switch (len) {
	case 11:
		c += ((unsigned long) key_p[10] << 24);
	case 10:
		c += ((unsigned long) key_p[9] << 16);
	case 9:
		c += ((unsigned long) key_p[8] << 8);
		/* the first byte of c is reserved for the length */
	case 8:
		b += ((unsigned long) key_p[7] << 24);
	case 7:
		b += ((unsigned long) key_p[6] << 16);
	case 6:
		b += ((unsigned long) key_p[5] << 8);
	case 5:
		b += key_p[4];
	case 4:
		a += ((unsigned long) key_p[3] << 24);
	case 3:
		a += ((unsigned long) key_p[2] << 16);
	case 2:
		a += ((unsigned long) key_p[1] << 8);
	case 1:
		a += key_p[0];
		/* case 0: nothing left to add */
	}
	HASH_MIX(a, b, c);

	return c;
}

void storage_die()
{
	exit(0);
}

/*  _________________________________________________________________
**
**  Support for Global Configuration
**  _________________________________________________________________
*/
#ifdef EMBED_LUA
#define STORAGE_MOD_CONFIG_KEY "lua_module"
#else
#define STORAGE_MOD_CONFIG_KEY "storage_module"
#endif

void storage_init_Child(apr_pool_t *p, server_rec *s)
{
	LUAConfigRecord *mc = ThisConfigRecord(s);

	mc->pid = getpid(); /* only call getpid() once per-process */

	/* XXX: there should be an ap_srand() function */
	srand((unsigned int)time(NULL));

	/* open the mutex lockfile */
	storage_mutex_reinit(s, p);
}

LUAConfigRecord *storage_config_global_create(server_rec *s)
{
	apr_pool_t *pool = s->process->pool;
	LUAConfigRecord *mc;
	void *vmc;

	apr_pool_userdata_get(&vmc, STORAGE_MOD_CONFIG_KEY, pool);
	if (vmc) {
		return vmc; /* reused for lifetime of the server */
	}

	/*
	* allocate an own subpool which survives server restarts
	*/
	mc = (LUAConfigRecord *)apr_palloc(pool, sizeof(*mc));
	mc->pPool = pool;
	mc->bFixed = FALSE;

	/*
	* initialize per-module configuration
	*/
	mc->nStorageMode      = STORAGE_SCMODE_UNSET;
	mc->szStorageDataFile = NULL;
	mc->nStorageDataSize  = 0;
	mc->pStorageDataMM    = NULL;
	mc->pStorageDataRMM   = NULL;
	mc->tStorageDataTable = NULL;
	mc->nMutexMode             = STORAGE_MUTEXMODE_UNSET;
	mc->nMutexMech             = APR_LOCK_DEFAULT;
	mc->szMutexFile            = NULL;
	mc->pMutex                 = NULL;

	apr_pool_userdata_set(mc, STORAGE_MOD_CONFIG_KEY,
		apr_pool_cleanup_null,
		pool);

	return mc;
}

static LUASrvConfigRec *storage_config_server_new(apr_pool_t *p)
{
	LUASrvConfigRec *sc = apr_palloc(p, sizeof(*sc));

	sc->mc                     = NULL;
	sc->enabled                = UNSET;
	sc->storage_timeout        = UNSET;
	sc->release                = FALSE;

	sc->init_script = NULL;
	sc->main_script = NULL;
	sc->http_init = NULL;
	sc->access_check = NULL;
	sc->check_user_id = NULL;
	sc->storage_timeout = 600;

	sc->L					   = NULL;
#ifdef APR_HAS_THREADS
	sc->is_threaded_mpm        = ap_find_linked_module("prefork.c") == NULL;
	if(sc->is_threaded_mpm)
	{
		sc->process_queue          = NULL;
		sc->process_queue_length   = 16;
	}else
	{
		sc->process_queue          = NULL;
		sc->process_queue_length   = 0;
	}
#endif
	return sc;
}

void storage_config_global_fix(LUAConfigRecord *mc)
{
	mc->bFixed = TRUE;
}

BOOL storage_config_global_isfixed(LUAConfigRecord *mc)
{
	return mc->bFixed;
}

void *storage_config_server_create(apr_pool_t *p, server_rec *s)
{
	LUASrvConfigRec *sc = storage_config_server_new(p);
	sc->mc = storage_config_global_create(s);
	return sc;
}
/*
*  Merge per-server STORAGE configurations
*/
#define cfgMerge(el,unset)  mrg->el = (add->el == (unset)) ? base->el : add->el
#define cfgMergeArray(el)   mrg->el = apr_array_append(p, add->el, base->el)
#define cfgMergeString(el)  cfgMerge(el, NULL)
#define cfgMergeBool(el)    cfgMerge(el, UNSET)
#define cfgMergeInt(el)     cfgMerge(el, UNSET)

void *storage_config_server_merge(apr_pool_t *p, void *basev, void *addv)
{
	LUASrvConfigRec *base = (LUASrvConfigRec *)basev;
	LUASrvConfigRec *add  = (LUASrvConfigRec *)addv;
	LUASrvConfigRec *mrg  = storage_config_server_new(p);

	cfgMerge(mc, NULL);
	cfgMergeBool(enabled);
	cfgMergeInt(storage_timeout);

	return mrg;
}
/*
*  Per-module initialization
*/
int storage_init_Module(apr_pool_t *p, apr_pool_t *plog,
					apr_pool_t *ptemp,
					server_rec *base_server)
{
	LUAConfigRecord *mc = ThisConfigRecord(base_server);
	LUASrvConfigRec *sc;
	server_rec *s;

	/*
	* Let us cleanup on restarts and exists
	*/
	apr_pool_cleanup_register(p, base_server,
		storage_init_ModuleKill,
		apr_pool_cleanup_null);

	/*
	* Any init round fixes the global config
	*/
	storage_config_global_create(base_server); /* just to avoid problems */
	storage_config_global_fix(mc);

	/*
	*  try to fix the configuration and open the dedicated STORAGE
	*  logfile as early as possible
	*/
	for (s = base_server; s; s = s->next) {
		sc = ThisSrvConfig(s);

		/* Fix up stuff that may not have been set */
		if (sc->enabled == UNSET) {
			sc->enabled = FALSE;
		}

		if (sc->storage_timeout == UNSET) {
			sc->storage_timeout = STORAGE_CACHE_TIMEOUT;
		}
	}

	/*
	* initialize the mutex handling
	*/
	if (!storage_mutex_init(base_server, p)) {
		return HTTP_INTERNAL_SERVER_ERROR;
	}

	/*
	* initialize storage caching
	*/
	storage_init(base_server, p);

	/*
	*  initialize servers
	*/
	ap_log_error(APLOG_MARK, APLOG_INFO, 0, base_server,
		"Init: Initializing (virtual) servers for STORAGE");

#ifdef APR_HAS_THREADS
	for (s = base_server; s; s = s->next) 
	{
		apr_status_t status;
		sc = ThisSrvConfig(s);

		status = apr_queue_create(&(sc->process_queue), sc->process_queue_length, p);
		assert(status==APR_SUCCESS && sc->process_queue!=NULL);
		if(status!=APR_SUCCESS || sc->process_queue==NULL)
		{
			ap_log_error(APLOG_MARK, APLOG_CRIT | APLOG_NOERRNO, status, s,
				"failed to create task queue");
			return status;
		}

		/*
		* Either now skip this server when STORAGE is disabled for
		* it or give out some information about what we're
		* configuring.
		*/
	}
#endif
	return OK;
}

apr_status_t storage_init_ModuleKill(void *data)
{
	server_rec *base_server = (server_rec *)data;
	server_rec *s;

	/*
	* Drop the storage cache and mutex
	*/
	storage_kill(base_server);

#ifdef APR_HAS_THREADS
	/*
	* Free the non-pool allocated structures
	* in the per-server configurations
	*/
	for (s = base_server; s; s = s->next) 
	{
		LUASrvConfigRec *sc = ThisSrvConfig(s);
		apr_status_t status;
		status = apr_queue_term(sc->process_queue);
	}
#endif

	return APR_SUCCESS;
}

/*  _________________________________________________________________
**
**  Session Cache: Common Abstraction Layer
**  _________________________________________________________________
*/

void storage_init(server_rec *s, apr_pool_t *p)
{
	LUAConfigRecord *mc = ThisConfigRecord(s);

	/*
	* Warn the user that he should use the storage cache.
	* But we can operate without it, of course.
	*/
	if (mc->nStorageMode == STORAGE_SCMODE_UNSET) {
		ap_log_error(APLOG_MARK, APLOG_WARNING, 0, s,
			"Init: Session Cache is not configured "
			"[hint: LUASessionCache]");
		mc->nStorageMode = STORAGE_SCMODE_NONE;
		return;
	}

	if (mc->nStorageMode == STORAGE_SCMODE_DBM)
		storage_dbm_init(s, p);
	else if (mc->nStorageMode == STORAGE_SCMODE_SHMHT
		||mc->nStorageMode == STORAGE_SCMODE_SHMCB) 
	{
		void *data;
		const char *userdata_key = "storage_init";

		apr_pool_userdata_get(&data, userdata_key, s->process->pool);
		if (!data) {
			apr_pool_userdata_set((const void *)1, userdata_key,
				apr_pool_cleanup_null, s->process->pool);
			return;
		}
		if (mc->nStorageMode == STORAGE_SCMODE_SHMHT)
			storage_shmht_init(s, p);
		else if (mc->nStorageMode == STORAGE_SCMODE_SHMCB)
			storage_shmcb_init(s, p);
	}
}

void storage_kill(server_rec *s)
{
	LUAConfigRecord *mc = ThisConfigRecord(s);

	if (mc->nStorageMode == STORAGE_SCMODE_DBM)
		storage_dbm_kill(s);
	else if (mc->nStorageMode == STORAGE_SCMODE_SHMHT)
		storage_shmht_kill(s);
	else if (mc->nStorageMode == STORAGE_SCMODE_SHMCB)
		storage_shmcb_kill(s);
	return;
}

BOOL storage_store(server_rec *s, UCHAR *id, int idlen, time_t expiry, void * pdata, int ndata)
{
	LUAConfigRecord *mc = ThisConfigRecord(s);
	BOOL rv = FALSE;

	if (mc->nStorageMode == STORAGE_SCMODE_DBM)
		rv = storage_dbm_store(s, id, idlen, expiry, pdata, ndata);
	else if (mc->nStorageMode == STORAGE_SCMODE_SHMHT)
		rv = storage_shmht_store(s, id, idlen, expiry, pdata, ndata);
	else if (mc->nStorageMode == STORAGE_SCMODE_SHMCB)
		rv = storage_shmcb_store(s, id, idlen, expiry, pdata, ndata);
	return rv;
}

void *storage_retrieve(server_rec *s, UCHAR *id, int idlen, int* ndata)
{
	LUAConfigRecord *mc = ThisConfigRecord(s);
	void *pdata = NULL;

	if (mc->nStorageMode == STORAGE_SCMODE_DBM)
		pdata = storage_dbm_retrieve(s, id, idlen,ndata);
	else if (mc->nStorageMode == STORAGE_SCMODE_SHMHT)
		pdata = storage_shmht_retrieve(s, id, idlen,ndata);
	else if (mc->nStorageMode == STORAGE_SCMODE_SHMCB)
		pdata = storage_shmcb_retrieve(s, id, idlen,ndata);
	return pdata;
}

void storage_remove(server_rec *s, UCHAR *id, int idlen)
{
	LUAConfigRecord *mc = ThisConfigRecord(s);

	if (mc->nStorageMode == STORAGE_SCMODE_DBM)
		storage_dbm_remove(s, id, idlen);
	else if (mc->nStorageMode == STORAGE_SCMODE_SHMHT)
		storage_shmht_remove(s, id, idlen);
	else if (mc->nStorageMode == STORAGE_SCMODE_SHMCB)
		storage_shmcb_remove(s, id, idlen);
	return;
}

void storage_status(server_rec *s, apr_pool_t *p, void (*func)(char *, void *), void *arg)
{
	LUAConfigRecord *mc = ThisConfigRecord(s);

	if (mc->nStorageMode == STORAGE_SCMODE_DBM)
		storage_dbm_status(s, p, func, arg);
	else if (mc->nStorageMode == STORAGE_SCMODE_SHMHT)
		storage_shmht_status(s, p, func, arg);
	else if (mc->nStorageMode == STORAGE_SCMODE_SHMCB)
		storage_shmcb_status(s, p, func, arg);

	return;
}

void storage_expire(server_rec *s)
{
	LUAConfigRecord *mc = ThisConfigRecord(s);

	if (mc->nStorageMode == STORAGE_SCMODE_DBM)
		storage_dbm_expire(s);
	else if (mc->nStorageMode == STORAGE_SCMODE_SHMHT)
		storage_shmht_expire(s);
	else if (mc->nStorageMode == STORAGE_SCMODE_SHMCB)
		storage_shmcb_expire(s);
	return;
}


/*
*  Configuration functions for particular directives
*/

const char *lua_cmd_Mutex(cmd_parms *cmd,
							 void *dcfg,
							 const char *arg)
{
	const char *err;
	LUAConfigRecord *mc = ThisConfigRecord(cmd->server);

	if ((err = ap_check_cmd_context(cmd, GLOBAL_ONLY))) {
		return err;
	}

	if (storage_config_global_isfixed(mc)) {
		return NULL;
	}

	if (strcEQ(arg, "none") || strcEQ(arg, "no")) {
		mc->nMutexMode  = STORAGE_MUTEXMODE_NONE;
	}
	/* NOTE: previously, 'yes' implied 'sem' */
	else if (strcEQ(arg, "default") || strcEQ(arg, "yes")) {
		mc->nMutexMode  = STORAGE_MUTEXMODE_USED;
		mc->nMutexMech = APR_LOCK_DEFAULT;
		mc->szMutexFile = NULL; /* APR determines temporary filename */
	}
#if APR_HAS_FLOCK_SERIALIZE
	else if (strlen(arg) > 6 && strcEQn(arg, "flock:", 6)) {
		const char *file = ap_server_root_relative(cmd->pool, arg+6);
		if (!file) {
			return apr_pstrcat(cmd->pool, "Invalid LUAMutex flock: path ", 
				arg+6, NULL);
		}
		mc->nMutexMode  = STORAGE_MUTEXMODE_USED;
		mc->nMutexMech = APR_LOCK_FLOCK;
		mc->szMutexFile = apr_psprintf(mc->pPool, "%s.%lu",
			file, (unsigned long)getpid());
	}
#endif
#if APR_HAS_FCNTL_SERIALIZE
	else if (strlen(arg) > 6 && strcEQn(arg, "fcntl:", 6)) {
		const char *file = ap_server_root_relative(cmd->pool, arg+6);
		if (!file) {
			return apr_pstrcat(cmd->pool, "Invalid LUAMutex fcntl: path ", 
				arg+6, NULL);
		}
		mc->nMutexMode  = STORAGE_MUTEXMODE_USED;
		mc->nMutexMech = APR_LOCK_FCNTL;
		mc->szMutexFile = apr_psprintf(mc->pPool, "%s.%lu",
			file, (unsigned long)getpid());
	}
#endif
#if APR_HAS_SYSVSEM_SERIALIZE && !defined(PERCHILD_MPM)
	else if (strcEQ(arg, "sysvsem")) {
		mc->nMutexMode  = STORAGE_MUTEXMODE_USED;
		mc->nMutexMech = APR_LOCK_SYSVSEM;
		mc->szMutexFile = NULL; /* APR determines temporary filename */
	}
#endif
#if APR_HAS_POSIXSEM_SERIALIZE
	else if (strcEQ(arg, "posixsem")) {
		mc->nMutexMode  = STORAGE_MUTEXMODE_USED;
		mc->nMutexMech = APR_LOCK_POSIXSEM;
		mc->szMutexFile = NULL; /* APR determines temporary filename */
	}
#endif
#if APR_HAS_PROC_PTHREAD_SERIALIZE
	else if (strcEQ(arg, "pthread")) {
		mc->nMutexMode  = STORAGE_MUTEXMODE_USED;
		mc->nMutexMech = APR_LOCK_PROC_PTHREAD;
		mc->szMutexFile = NULL; /* APR determines temporary filename */
	}
#endif
#if APR_HAS_FLOCK_SERIALIZE || APR_HAS_FCNTL_SERIALIZE
	else if (strlen(arg) > 5 && strcEQn(arg, "file:", 5)) {
		const char *file = ap_server_root_relative(cmd->pool, arg+5);
		if (!file) {
			return apr_pstrcat(cmd->pool, "Invalid LUAMutex file: path ", 
				arg+5, NULL);
		}
		mc->nMutexMode  = STORAGE_MUTEXMODE_USED;
#if APR_HAS_FLOCK_SERIALIZE
		mc->nMutexMech  = APR_LOCK_FLOCK;
#endif
#if APR_HAS_FCNTL_SERIALIZE
		mc->nMutexMech  = APR_LOCK_FCNTL;
#endif
		mc->szMutexFile =
			apr_psprintf(mc->pPool, "%s.%lu",
			file, (unsigned long)getpid());
	}
#endif
#if (APR_HAS_SYSVSEM_SERIALIZE && !defined(PERCHILD_MPM)) || APR_HAS_POSIXSEM_SERIALIZE
	else if (strcEQ(arg, "sem")) {
		mc->nMutexMode  = STORAGE_MUTEXMODE_USED;
#if APR_HAS_SYSVSEM_SERIALIZE && !defined(PERCHILD_MPM)
		mc->nMutexMech  = APR_LOCK_SYSVSEM;
#endif
#if APR_HAS_POSIXSEM_SERIALIZE
		mc->nMutexMech  = APR_LOCK_POSIXSEM;
#endif
		mc->szMutexFile = NULL; /* APR determines temporary filename */
	}
#endif
	else {
		return apr_pstrcat(cmd->pool, "Invalid LUAMutex argument ", 
			arg, " (", lua_valid_mutex_string, ")", NULL);
	}

	return NULL;
}

#define MODLUA_NO_SHARED_MEMORY_ERROR \
	"LUAStorage: shared memory cache not useable on this platform"

const char *lua_cmd_Release(cmd_parms *cmd, void *dummy, int arg)
{
	LUASrvConfigRec *sc = ThisSrvConfig(cmd->server);
	sc->release = arg;
	return NULL;
}

const char *lua_cmd_process_queue_Length(cmd_parms *cmd, void *dcfg, const char *arg)
{
	LUASrvConfigRec *sc = ThisSrvConfig(cmd->server);

#ifdef APR_HAS_THREADS
	if (sc->is_threaded_mpm)
	{
		sc->process_queue_length = atoi(arg);

		if (sc->process_queue_length <= 0 || sc->process_queue_length > 256) {
			return "LUA_process_queue_Length: Invalid argument, must be greater than 0 or less than 256";
		}
		return NULL;
	}else
		return "LUA_process_queue_Length only available in thread mpm! ";
#else
	return "LUA_process_queue_Length only available in thread mpm! ";
#endif
}
const char *lua_cmd_Storage(cmd_parms *cmd,
									void *dcfg,
									const char *arg)
{
	LUAConfigRecord *mc = ThisConfigRecord(cmd->server);

	const char *err, *colon;
	char *cp, *cp2;
	int arglen = strlen(arg);

	if ((err = ap_check_cmd_context(cmd, GLOBAL_ONLY))) {
		return err;
	}

	if (storage_config_global_isfixed(mc)) {
		return NULL;
	}

	if (strcEQ(arg, "none")) {
		mc->nStorageMode      = STORAGE_SCMODE_NONE;
		mc->szStorageDataFile = NULL;
	}
	else if ((arglen > 4) && strcEQn(arg, "dbm:", 4)) {
		mc->nStorageMode      = STORAGE_SCMODE_DBM;
		mc->szStorageDataFile = ap_server_root_relative(mc->pPool, arg+4);
		if (!mc->szStorageDataFile) {
			return apr_psprintf(cmd->pool,
				"LUAStorage: Invalid cache file path %s",
				arg+4);
		}
	}
	else if ((arglen > 6) && (strcEQn(arg, "shmht:", 6) || strcEQn(arg, "shmcb:", 6)) ) {
#if !APR_HAS_SHARED_MEMORY
		return MODLUA_NO_SHARED_MEMORY_ERROR;
#endif
		if(strcEQn(arg, "shmht:", 6))
			mc->nStorageMode = STORAGE_SCMODE_SHMHT;
		if(strcEQn(arg, "shmcb:", 6))
			mc->nStorageMode = STORAGE_SCMODE_SHMCB;

		colon = ap_strchr_c(arg, ':');
		mc->szStorageDataFile =
			ap_server_root_relative(mc->pPool, colon+1);
		if (!mc->szStorageDataFile) {
			return apr_psprintf(cmd->pool,
				"LUAStorage: Invalid storage file path %s",
				colon+1);
		}
		mc->tStorageDataTable = NULL;
		mc->nStorageDataSize  = 1024*512; /* 512KB */

		if ((cp = strchr(mc->szStorageDataFile, '('))) {
			*cp++ = NUL;

			if (!(cp2 = strchr(cp, ')'))) {
				return "LUAStorage: Invalid argument: "
					"no closing parenthesis";
			}

			*cp2 = NUL;

			mc->nStorageDataSize = atoi(cp);

			if (mc->nStorageDataSize < 8192) {
				return "LUAStorage: Invalid argument: "
					"size has to be >= 8192 bytes";
			}

			if (mc->nStorageDataSize >= APR_SHM_MAXSIZE) {
				return apr_psprintf(cmd->pool,
					"LUAStorage: Invalid argument: "
					"size has to be < %d bytes on this "
					"platform", APR_SHM_MAXSIZE);
			}
		}
	}
	else {
		return "LUAStorage: Invalid argument,valid argument like (none,shmht:logs/scache(512000), dbm:logs/scache) ";
	}

	return NULL;
}

const char *lua_cmd_Http(cmd_parms *cmd,
						 void *dcfg,
						 const char *arg)
{
	LUASrvConfigRec *sc = ThisSrvConfig(cmd->server);
	sc->http_init = (char *)arg;

	return NULL;
}

const char *lua_cmd_Init(cmd_parms *cmd,
										   void *dcfg,
										   const char *arg)
{
	LUASrvConfigRec *sc = ThisSrvConfig(cmd->server);
	sc->init_script = (char *)arg;

	return NULL;
}

const char *lua_cmd_Main(cmd_parms *cmd,
										   void *dcfg,
										   const char *arg)
{
	LUASrvConfigRec *sc = ThisSrvConfig(cmd->server);
	sc->main_script = (char *)arg;

	return NULL;
}

const char *lua_cmd_Handle(cmd_parms *cmd,
						   void *dcfg,
						   const char *arg)
{
	ml_directory *cfg = (ml_directory *) dcfg;
	(void)cmd; /* avoid "unused parameter" warning */
	cfg->handlers = (char *)arg;
	return NULL;
}

const char *lua_cmd_Access(cmd_parms *cmd,
						   void *dcfg,
						   const char* arg)
{
	LUASrvConfigRec *sc = ThisSrvConfig(cmd->server);

	sc->access_check = (char *)arg;

	return NULL;
}

const char *lua_cmd_Check_user_id(cmd_parms *cmd,
						   void *dcfg,
						   const char* arg)
{
	LUASrvConfigRec *sc = ThisSrvConfig(cmd->server);

	sc->check_user_id = (char *)arg;

	return NULL;
}



const char *lua_cmd_StorageTimeout(cmd_parms *cmd,
										   void *dcfg,
										   const char *arg)
{
	LUASrvConfigRec *sc = ThisSrvConfig(cmd->server);

	sc->storage_timeout = atoi(arg);

	if (sc->storage_timeout <= 0) {
		return "LUAStorageTimeout: Invalid argument, must be greater than 0";
	}

	return NULL;
}

/*  _________________________________________________________________
**
**  STORAGE Extension to mod_status
**  _________________________________________________________________
*/
#if 0 /* NOT YET */
static void storage_ext_ms_display(request_rec *, int, int);

void storage_status_register(apr_pool_t *p)
{
	/* XXX point mod_status to this update, when it grows the opt fn */
#if 0
	ap_hook_register("ap::mod_status::display", storage_ext_ms_display, AP_HOOK_NOCTX);
#endif
	return;
}

static void storage_ext_ms_display_cb(char *str, void *_r)
{
	request_rec *r = (request_rec *)_r;
	if (str != NULL)
		ap_rputs(str, r);
	return;
}

static void storage_ext_ms_display(request_rec *r, int no_table_report, int short_report)
{
	LUASrvConfigRec *sc = ThisSrvConfig(cmd->server);

	if (sc == NULL)
		return;
	if (short_report)
		return;
	ap_rputs("<hr>\n", r);
	ap_rputs("<table cellspacing=0 cellpadding=0>\n", r);
	ap_rputs("<tr><td bgcolor=\"#000000\">\n", r);
	ap_rputs("<b><font color=\"#ffffff\" face=\"Arial,Helvetica\">LUA Session Cache Status:</font></b>\r", r);
	ap_rputs("</td></tr>\n", r);
	ap_rputs("<tr><td bgcolor=\"#ffffff\">\n", r);
	storage_status(r->server, r->pool, storage_ext_ms_display_cb, r);
	ap_rputs("</td></tr>\n", r);
	ap_rputs("</table>\n", r);
	return;
}
#endif
